﻿<#
    Script om volgende stappen uit te voeren ten behoeve van het Benzo-GGZ project (en andere medicatie-indicatoren)
    1. Check voor aanwezigheid van deze bestanden:
        patientverblijven.csv en toedieningen.csv
    2. Lees deze CSV-bestanden in
    3. Pseudonimiseer en schrijf de correspondentietabel (pseudoniemen.CSV)
    4. schrijf de data weg als een XML-bestand
    5. bereken de hash.
    6. Geef de boodschap om deze bestanden op de server te plaatsen (NIET correspondentietabel)

    Auteur: Dirk De Wachter
    Datum: 15 februari 2021
    Versie: Test 0.5 - met WinForms, datafile, afdelingen en nieuwe definities van de CSV-bestanden.
        Test 0.8 - Met toevoeging van volgnummer afdeling
        Test 0.9 - Met aparte thread voor verwerking
        Test 0.91 - Met bijkomende test voor de variabelen in de CSV-bestanden
#>

# Load the assembly
try {
    Add-Type -AssemblyName System.Windows.Forms | Out-Null
    Add-Type -AssemblyName System.Drawing
    [System.Windows.Forms.Application]::EnableVisualStyles()
} catch {
    Throw "De nodige Windows module(s) kunnen niet geladen worden. Contacteer VIKZ."
}

#region Data
<# Initialisatie en definities #>
$FullPath=$PSCommandPath # $script:MyInvocation.MyCommand.Path
#Write-Host $PSCommandPath
$CDir=$PSScriptRoot # $(Split-Path $FullPath) #Get-Location
#Write-Host $PSScriptRoot
# Of gebruik een referentiebestand Voorzieningen.psd1 en import met Import-PowerShellDataFile
$_Exist = Get-ChildItem -Path "Filesystem::$CDir\*.psd1" -File -Include voorzieningen.psd1
If ($_Exist.Count -lt 1) { 
    New-MessageBox -message "Het bestand ""voorzieningen.psd1"" ontbreekt in deze folder ($CDir). Gelieve de software opnieuw volledig uit te pakken. Het script wordt afgesloten." `
            -title 'Ontbrekend bestand' -buttons 'OK' 
    break
}
$UserIds=(Import-PowerShellDataFile $CDir\voorzieningen.psd1)
$script:bestanden=@("patientverblijven.csv","toedieningen.csv")

<# Open een logbestand #>
$LogFile = $FullPath -replace "ps1$","log"
Get-Date | Out-File -Encoding utf8 -FilePath $Logfile
#endregion Data

#region Forms
<# functies #>
$DefMsgBox = {
function New-MessageBox {
    param (
        [string]$Message = '',
        [string]$Title = 'Message box title',   
        [string]$buttons = 'OKCancel'
    )
    # This function displays a message box by calling the .Net Windows.Forms (MessageBox class)
 
 
    # Define the button types
    switch ($buttons) {
       'ok' {$btn = [System.Windows.Forms.MessageBoxButtons]::OK; break}
       'okcancel' {$btn = [System.Windows.Forms.MessageBoxButtons]::OKCancel; break}
       'AbortRetryIgnore' {$btn = [System.Windows.Forms.MessageBoxButtons]::AbortRetryIgnore; break}
       'YesNoCancel' {$btn = [System.Windows.Forms.MessageBoxButtons]::YesNoCancel; break}
       'YesNo' {$btn = [System.Windows.Forms.MessageBoxButtons]::yesno; break}
       'RetryCancel'{$btn = [System.Windows.Forms.MessageBoxButtons]::RetryCancel; break}
       default {$btn = [System.Windows.Forms.MessageBoxButtons]::RetryCancel; break}
    }
    # Display the message box
    Return ([System.Windows.Forms.MessageBox]::Show($Message,$Title,$btn))
}
}

Function New-FormControl {
param (
    [string]$ControlType,
    [string]$Name,
    [System.Windows.Forms.Control]$Parent,
    [int[]]$Size,
    [int[]]$Location,
    [string[]]$Anchor,
    [string]$Text,
    [string[]]$Font
)
    switch($ControlType) {
        'Form'        {$FC = New-Object System.Windows.Forms.Form; break }
        'Listbox'     {$FC = New-Object System.Windows.Forms.Listbox; break }
        'ComboBox'    {$FC = New-Object System.Windows.Forms.ComboBox; break }
        'CheckBox'    {$FC = New-Object System.Windows.Forms.CheckBox; break }
        'ProgressBar' {$FC = New-Object System.Windows.Forms.ProgressBar; break }
        'TextBox'     {$FC = New-Object System.Windows.Forms.TextBox; break }
        'Button'      {$FC = New-Object System.Windows.Forms.Button; break }
        'Label'       {$FC = New-Object System.Windows.Forms.Label; break }
        default       {$FC = New-Object System.Windows.Forms.Label; break }
    }
    $FC.Name=$Name
    If ($ControlType -eq "Form") { $FC.ClientSize = New-Object System.Drawing.Point($Size[0], $Size[1]) }
    Else { $FC.Size = New-Object System.Drawing.Size($Size[0], $Size[1]) }
    If ($Location) { $FC.location = New-Object System.Drawing.Point($Location[0],$Location[1]) }
    If ($Anchor) { $FC.Anchor = $Anchor }
    If ($Text -ne "") { $FC.text = $Text }
    If ($Font) { $FC.Font = New-Object System.Drawing.Font($Font) }
    If ($null -ne $Parent) { $Parent.Controls.Add($FC) }
    Return $FC
}

Function MaakForm {

<# This form was prepared using POSHGUI.com a free online gui designer for PowerShell #>
$BenzoForm                       = New-FormControl -ControlType "Form" -Name "BenzoForm" -Size 500,333 -Text "Benzo data conversie"
$BenzoForm.StartPosition         = [System.Windows.Forms.FormStartPosition]::CenterScreen
$BenzoForm.TopMost               = $true
$BenzoForm.MaximizeBox           = $false
$ZiekenhuisLabel                 = New-FormControl -ControlType "Label" -Name "Ziekenhuislabel" -Parent $BenzoForm -Size 25,10 -Location 12,11 -Text "Kies je ziekenhuis" -Font "Microsoft Sans Serif",10
$ZiekenhuisLabel.AutoSize        = $true
$Ziekenhuis                      = New-FormControl -ControlType "ListBox" -Name "Ziekenhuis" -Parent $BenzoForm -Size 320,96 -Location 12,28 
$Registratiejaar                 = New-FormControl -ControlType "ComboBox" -Name "Registratiejaar" -Parent $BenzoForm -Size 120,20 -Location 352,13 -Anchor "top,right" -Text "Registratiejaar" -Font "Microsoft Sans Serif",10
$ProgressBar                     = New-FormControl -ControlType "ProgressBar" -Name "ProgressBar" -Parent $BenzoForm -Size 460,19 -Location 12,296 -Anchor "right,bottom,left" -Text "Voortgang"
$ProgressBar.BackColor           = [System.Drawing.ColorTranslator]::FromHtml("#7ed321")
$ProgressBar.Minimum             = 0
$ProgressBar.Maximum             = 100 # Pct
$ProgressBar.Step                = 1
$StatusLabel                     = New-FormControl -ControlType "Label" -Name "StatusLabel" -Parent $BenzoForm -Size 25,10 -Location 12,172 -Text "Status" -Font "Microsoft Sans Serif",10
$StatusLabel.AutoSize            = $true
$Status                          = New-FormControl -ControlType "TextBox" -Name "Status" -Parent $BenzoForm -Size 460,98 -Location 12,190 -Anchor "top,right,left" -Font "Microsoft Sans Serif",10
$Status.multiline                = $true
$Status.Enabled                  = $true
$Status.ReadOnly                 = $true
$Status.ScrollBars               = 2 # Vertical
$Continue                        = New-FormControl -ControlType "Button" -Name "Continue" -Parent $BenzoForm -Size 120,30 -Location 352,93 -Anchor "top,right" -Text "Verwerk" -Font "Microsoft Sans Serif",10
$Continue.enabled                = $true
$EmailLabel                      = New-FormControl -ControlType "Label" -Name "EmailLabel" -Parent $BenzoForm -Size 25,10 -Location 352,46 -Text "Emailadres" -Font "Microsoft Sans Serif",10
$EmailLabel.AutoSize             = $true
$Email                           = New-FormControl -ControlType "TextBox" -Name "Email" -Parent $BenzoForm -Size 120,20 -Location 352,66 -Anchor "top,right" -Font "Microsoft Sans Serif",10
$Email.multiline                 = $false
$patientverblijven               = New-FormControl -ControlType "CheckBox" -Name "patientverblijven" -Parent $BenzoForm -Size 415,20 -Location 56,138 -Anchor "top,left" -Text $bestanden[0] -Font "Microsoft Sans Serif",10
$patientverblijven.AutoSize      = $false
$patientverblijven.BackColor     = [System.Drawing.ColorTranslator]::FromHtml("#dd606e") # = niet gevonden
$patientverblijven.ThreeState    = $true
$patientverblijven.Enabled       = $false
$patientverblijven.CheckState    = 2 # Indeterminate
$toedieningen                    = New-FormControl -ControlType "CheckBox" -Name "toedieningen" -Parent $BenzoForm -Size 415,20 -Location 56,156 -Anchor "top,left" -Text $bestanden[1] -Font "Microsoft Sans Serif",10
$toedieningen.AutoSize           = $false
$toedieningen.BackColor          = [System.Drawing.ColorTranslator]::FromHtml("#dd606e") # = niet gevonden
$toedieningen.ThreeState         = $true
$toedieningen.Enabled            = $false
$toedieningen.CheckState         = 2 # Indeterminate

<# Initialize #>
$script:DefaultColor=$Status.BackColor
##$Ziekenhuis.DataBindings.Add("Items",($UserIds.GetEnumerator()),"Value")
<#$UserIds.GetEnumerator() | Select-Object -Property Value  | Sort-Object -Property Value |`
    ForEach-Object {[void] $Ziekenhuis.Items.Add($_.Value)}#>
$Ziekenhuis.Items.AddRange(($UserIds.GetEnumerator() | Sort-Object -Property Value | Select-Object -ExpandProperty Value  ))
$script:userid = @{ "0"="" } # leeg
$CurrentYear = Get-Date -UFormat %Y
#For ($y=-1; $y -gt -4; $y--) { [void] $Registratiejaar.Items.Add([Int]$CurrentYear + $y) }
$Registratiejaar.Items.AddRange(@(([int]$CurrentYear-1),([int]$CurrentYear-2),([int]$CurrentYear-3)))
$script:regjaar = ""
$Status.Text = "Welkom "+$env:USERNAME+"." | Out-File $logfile -Encoding utf8 -Append
"U gebruikt Powershell versie " + ($PSVersionTable.PSVersion -join ".") | Out-File $logfile -Encoding utf8 -Append

# Check voor bestanden en pas checkboxes aan.
<# Stap 1 - Check voor bestanden #>
# Write-Host "Zijn alle bestanden aanwezig?"
$FL = Get-ChildItem -Path "Filesystem::$CDir\*.csv" -File -Include $bestanden | Sort-Object -Property Name | Select-Object -ExpandProperty Name
$bestandenGevonden=$false
If ($FL.Count -eq $bestanden.Length) {
    If ($FL[0].ToLower().Equals($bestanden[0].ToLower()) <#patientverblijven.csv#>) {
        $patientverblijven.CheckState=1 # Checked
        $patientverblijven.BackColor=[System.Drawing.ColorTranslator]::FromHtml("#b8e986")
    }
    else { $patientverblijven.CheckState=0 } # Unchecked
    If ($FL[1].ToLower().Equals($bestanden[1].ToLower()) <#toedieningen.csv#>) {
        $toedieningen.CheckState=1 # Checked
        $toedieningen.BackColor=[System.Drawing.ColorTranslator]::FromHtml("#b8e986")
    }
    else { $toedieningen.CheckState=0 } # Unchecked
    If ($patientverblijven.CheckState -eq 1 -and $toedieningen.CheckState -eq 1) {
        $Status.Text += "`r`nAlle bestanden gevonden."
        "Alle bestanden gevonden." | Out-File $logfile -Encoding utf8 -Append
        $bestandenGevonden=$true
    }
}
If ((-not $bestandenGevonden) -or ($FL.Count -ne $bestanden.Length)) {
    $message="De bestanden " + ($bestanden -join " en ") + " werden niet allen gevonden in deze folder ($CDir). Het script wordt afgesloten."
    New-MessageBox -message $message -title 'Onvolledige dataset' -buttons 'OK' 
    break
}

<# Events #>
$Registratiejaar.Add_SelectedValueChanged({ SelectJaar $_ })
$Ziekenhuis.Add_SelectedValueChanged({ SelectZiekenhuis $_ })
$Continue.Add_Click({ VoerUit $_ })

$BenzoForm.Add_Load({ 
    Write-Host "Benzo is geladen" 
})
$BenzoForm.Add_Shown({ 
    Write-Host "Benzo wordt getoond" 
    # Hide PowerShell Console
    Add-Type -Name Window -Namespace Console -MemberDefinition '
        [DllImport("Kernel32.dll")]
        public static extern IntPtr GetConsoleWindow();
        [DllImport("user32.dll")]
        public static extern bool ShowWindow(IntPtr hWnd, Int32 nCmdShow);
    '
    $consolePtr = [Console.Window]::GetConsoleWindow()
    [Console.Window]::ShowWindow($consolePtr, 0) | Out-Null
})

<# show window #>
[void]$BenzoForm.ShowDialog()
}


function SelectJaar {
    $Status.BackColor=$script:DefaultColor
    $Status.AppendText("`r`nRegistratiejaar " + $RegistratieJaar.SelectedItem + " is gekozen.")
    "Registratiejaar " + $RegistratieJaar.SelectedItem + " is gekozen." | Out-File $logfile -Encoding utf8 -Append
    $script:regjaar=$RegistratieJaar.SelectedItem
}

function SelectZiekenhuis {
    $Status.BackColor=$script:DefaultColor
    $Status.AppendText("`r`nZiekenhuis " + $Ziekenhuis.SelectedItem + " is gekozen.")
    "Ziekenhuis " + $Ziekenhuis.SelectedItem + " is gekozen."| Out-File $logfile -Encoding utf8 -Append
    $script:userid = $UserIds.GetEnumerator() | Where-Object { $($_.Value) -eq $Ziekenhuis.SelectedItem }
}
#endregion Forms

function VoerUit {
        $Compleet=1;
        $Status.BackColor=$script:DefaultColor
        If (!($Email.Text -as [Net.Mail.MailAddress])) { 
            $Status.AppendText("`r`nGeef een geldig emailadres op." )
            $Compleet=0
        }
        If ($script:userid.Name.length -lt 4) {
            $Status.AppendText("`r`nGeen voorziening gekozen.")
            $Compleet=0
        }
        If ($script:regjaar -eq "") {
            $Status.AppendText("`r`nGeen registratiejaar gekozen." )
            $Compleet=0
        }
        If ($Compleet) {
            $RegistratieJaar.Enabled = $false
            $Ziekenhuis.Enabled = $false
            $Email.Enabled = $false

            #Invoke-Expression -Command 
            #& "$VerwerkingsCode $Cdir $Logfile $Status $ProgressBar $regjaar $userid"
            $ps = (New-Runspace -Code $VerwerkingsCode -Arguments @($Cdir,$LogFile,$bestanden,$Status,$ProgressBar,$script:regjaar,$Email.Text,$script:userid))
            $job = $ps[0].BeginInvoke() # om op te starten ...
            While (-Not $job.IsCompleted) { Start-Sleep 2 }  # om te wachten tot ze afgelopen is ...
            $XMLFile = $ps[0].EndInvoke($job)
            Close-Runspace($ps[0])

            <# Stap #>
            #$XMLFile="$CDir\"+$userid.Name+"-benzo-"+$regjaar+"-"+$script:numpat+".xml"
            If ($XMLFile.Length -eq 0) {
                New-MessageBox -message "Er was een probleem met de databestanden. Bekijk het logbestand ($logfile) en controleer de gegevens."`
                    -title "Probleem databestand" -buttons 'OK'
            }
            Else {
                New-MessageBox -message "Plaats nu volgende bestanden op de sFTP-server`n$XMLFile`n$XMLFile.hash" `
                    -title 'Omzetting gegevens gelukt.' -buttons 'OK' 
            }
            Get-Date | Out-File -Encoding utf8 -FilePath $Logfile -Append
            $BenzoForm.Dispose()
        }
        Else { $Status.BackColor=[System.Drawing.ColorTranslator]::FromHtml("#ee8096") }
}

$VerwerkingsCode = {

<#variables#>
#output $logfile (+ $CDir -> XMLFile)
#GUI: $ProgressBar / $Status
#inputs: $regjaar $emailadres $userid
Param (
    [string]$Path,
    [string]$LogFile,
    [string[]]$bestanden,
    [System.Windows.Forms.Control]$Status,
    [System.Windows.Forms.Control]$ProgressBar,
    [string]$regjaar,
    [Net.Mail.MailAddress]$emailadres,
    [System.Collections.DictionaryEntry]$userid
)

If ($Path -eq "") { break }
$CDir=$Path

<# Arrays van de te selecteren variabelen in de bestanden #>
$script:verblijfNodig="Patient","Verblijf","Afdelingscode","TypeOpname","Opnamedatum","Ontslagdatum","MedischeOpname", "MedischOntslag","Geslacht","Geboortedatum"
$script:verblijfSelect="Patient","Verblijf","Volgorde","Afdelingscode","Afdeling","TypeOpname","Opnamedatum","Ontslagdatum","MedischeOpname", "MedischOntslag","Geslacht","Geboortedatum"
$script:toedieningNodig="Patient","Verblijf","Afdelingscode","Toedieningsdag","CNKUD","Aantal","Eenheid"
$script:toedieningSelect="Patient","Verblijf","Volgorde","Afdelingscode","Toedieningsdag","CNKUD","Aantal","Eenheid"
$script:afdelingSelect="Afdelingscode","Afdeling"

<# Stap 2 - Lees bestanden #>
function ControleerVariabelen([string]$bestand, [Object[]]$dataArray, [Object[]]$controleArray) {
    $gevondenVars=($dataArray | Get-Member -Type NoteProperty | Select-Object -ExpandProperty Name)
    $ontbrekendeVars = ($controleArray | Where-Object { $gevondenVars -notcontains $_})
    "`tGevonden variabelen in $bestand : "+($gevondenVars -join ", ") | Out-File $logfile -Encoding utf8 -Append

    If ($ontbrekendeVars.Length -gt 0) {
        $ontbrekendeLijst=($ontbrekendeVars -join ", ")
        $Status.AppendText("`r`n`tEr ontbreken variabelen in $bestand.")
        "`tIn het bestand $bestand ontbreken deze noodzakelijke variabelen: " + $ontbrekendeLijst +`
            ".`r`n`tGelieve het bestand te controleren. Het script wordt afgesloten." | Out-File $logfile -Encoding utf8 -Append
        return 1
    }
    return 0
}

function LeesBestand([Byte]$f, [Object[]]$Nodig, [Object[]]$Variabelen) {
    $b=$bestanden[$f]
    $CSVData=(Import-CSV -PATH $($CDir + "\"+$b) -Delimiter ";" )
    $Status.AppendText("`r`n`t$b is gelezen.")
    "`t$b is gelezen." | Out-File $logfile -Encoding utf8 -Append
    $r = ControleerVariabelen $b $CSVData $Nodig
    $processedData=$Null
    if ($r -eq 0) {
        $_l=1
        $processedData = ($CSVData | Select-Object -Property ($Variabelen + "Line"))
        $processedData | ForEach-Object { $_.Line=$_l++ }
        "`tAantal lijnen gelezen:"+ $processedData.Length | Out-File $logfile -Encoding utf8 -Append
    }
    return $processedData
}

function ControleerInhoud([Object[]]$Data, [String]$Variabele, [Object[]]$Inhoud) {
    $varInhoud=($Data | Select-Object -ExpandProperty $Variabele -Unique)
    $andereInhoud = ($varInhoud | Where-Object {$Inhoud -notcontains $_})
    If ($andereInhoud.Length -gt 0) {
        "`tDe variabele $Variabele bevat ongeldige inhoud: " + ($andereInhoud -join ",") | Out-File $logfile -Encoding utf8 -Append
        return $false 
    }
    return $true
}
function LeesBestanden {
    $Status.AppendText("`r`nStap 2:`tLees de csv-bestanden.")
    "Stap 2: Lees de csv-bestanden." | Out-File $logfile -Encoding utf8 -Append
    $script:patData=(LeesBestand 0 $script:verblijfNodig $script:verblijfSelect)
    $script:toedieningData=(LeesBestand 1 $script:toedieningNodig $script:toedieningSelect)

    <# controleer inhoud van variabelen #>
    $inhoud=$true
    $inhoud= ($inhoud -and $(ControleerInhoud $script:patData "geslacht" @("Vrouw","Man","")))
    $inhoud= ($inhoud -and $(ControleerInhoud $script:patData "TypeOpname" @(1,2,3,4,9)))
    $inhoud= ($inhoud -and $(ControleerInhoud $script:toedieningData "eenheid" @("MG","ML","CAPS","COMP","AMP")))
    if (-not $inhoud) { $Status.AppendText("`r`n`tOngeldige inhoud in databestanden.") }

    return ($inhoud -and ($Null -ne $script:patData) -and ($Null -ne $script:toedieningData))
}

<# Stap 3 - Pseudoniemen voor patienten #>

function MaakPseudoniemen {
$Status.AppendText("`r`nStap 3:`tGenereer pseudoniemen.")
"Stap 3:`tGenereer pseudoniemen." | Out-File $logfile -Encoding utf8 -Append
<# Maak een lijst van unieke patienten #>
$patients=$script:patData | Select-Object -Property $verblijfSelect[0] <# PID #> `
                | Sort-Object -Unique -Property $verblijfSelect[0]
<# Maak een tabel met random codes per patient #>
$script:pathash=[ordered]@{}
$script:numpat=$patients.length
$ProgressBar.Maximum=$script.numpat
$ProgressBar.Step=1
$ProgressBar.Value=0
ForEach ($pat in $patients) { 
    $G=New-Guid; 
    $pathash.Add((select-Object -ExpandProperty $verblijfSelect[0] -InputObject $pat),$G) 
    $ProgressBar.PerformStep()
}
$ProgressBar.Refresh() # catch up with progressbar.
$Status.AppendText("`r`n`tSchrijf pseudoniemen in pseudoniemen.csv.")
"`tSchrijf pseudoniemen in pseudoniemen.csv." | Out-File $logfile -Encoding utf8 -Append
$pathash.GetEnumerator() | Select-Object -Property Key,Value | Export-Csv -PATH $($CDir + "\pseudoniemen.csv") -UseCulture -NoTypeInformation
}

<# Stap 4 - Pseudoniemen voor patienten #>

Function Get-Datum ([string]$DatumTekst,[string[]]$errorTekst) {
    try {
        $Datum = [DateTime]::Parse($DatumTekst)
    }
    catch [System.FormatException] {
        $errorTekst | Out-File -FilePath $LogFile -Append # Write to logfile
        $Datum = $null
    }
    catch {
        $error | Out-File -FilePath $LogFile -Append # Write to logfile
        $Datum = $null
    }
    return $Datum
}

function MaakVolgnummers {
    $Status.AppendText("`r`nStap 4:`tBereken volgnummers.")
    "Stap 4:`tBereken volgnummers." | Out-File $logfile -Encoding utf8 -Append
    $lengte=$pathash.Count
    $ProgressBar.Maximum=$lengte
    $ProgressBar.Step=1
    $ProgressBar.Value=0
    ForEach ($pat in ($pathash.GetEnumerator() | Select-Object -Property Key)) {
        $records = ($patData | Where-Object { $_.Patient -eq $pat.Key } | `
            ForEach-Object { $_ | Add-Member -Name _medontslag `
                -Value (Get-Datum $_.MedischOntslag ("Ongeldige datum "+$_.MedischOntslag+" in $bestanden[0] op lijn "+($_.Line+1)+".")) `
                -MemberType NoteProperty -PassThru }  | `
            ForEach-Object { $_ | Add-Member -Name _opname `
                -Value (Get-Datum $_.opnamedatum ("Ongeldige datum "+$_.opnamedatum+" in $bestanden[0] op lijn "+($_.Line+1)+".")) `
                -MemberType NoteProperty -PassThru }  | `
            Sort-Object -Property _medontslag, _opname)
        if ($records -is [array]) {
            $o=0
            $v=0
            $last_mo = $records[0]._medontslag
            For ($r=0 ; $r -lt $records.Count; $r++) { 
                if ($records[$r]._medontslag -gt $last_mo) {
                    $o += 1
                    $v = 0
                    $last_mo = $records[$r]._medontslag
                }
                $records[$r] | Add-Member -name Verblijfsnummer -Value ($o+1) -MemberType NoteProperty -Force
                $records[$r] | Add-Member -Name Volgorde -Value ($v+1) -MemberType NoteProperty -Force
                if ($records[$r].ontslagdatum -eq "") { $_ontslag=Get-Date } 
                else {$_ontslag=(Get-Datum $records[$r].ontslagdatum ("Ongeldige datum "+$records[$r].ontslagdatum+" in $bestanden[0] op lijn "+($records[$r].Line+1)+".")) }
                $records[$r] | Add-Member -Name _ontslag -Value $_ontslag -MemberType NoteProperty
                $v++
            }
        }
        else {
                $records | Add-Member -name Verblijfsnummer -Value 1 -MemberType NoteProperty -Force
                $records | Add-Member -Name Volgorde -Value 1 -MemberType NoteProperty -Force
                if ($records.ontslagdatum -eq "") { $_ontslag=Get-Date } 
                else {$_ontslag=(Get-Datum $records.ontslagdatum ("Ongeldige datum "+$records.ontslagdatum+" in $bestanden[0] op lijn "+($records.Line+1)+".")) }
                $records | Add-Member -Name _ontslag -Value $_ontslag -MemberType NoteProperty
        }
        $ProgressBar.PerformStep()
    }
}


<# Stap 5 - hercodeer patData #>
function Get-Afdelingen {
    [CmdletBinding()]  # Add cmdlet features.
    param (
        [Parameter(ValueFromPipeline,Mandatory=$true)][Object[]]$Data
    )
    begin { 
        $afdelingen= @{}
        $alijst = @()
    }
    process { $alijst += ($Data | Select-Object -Property $afdelingSelect ) }
    end {
        $alijst | Sort-Object -Property $afdelingSelect -Unique | ForEach-Object { $afdelingen.Add($_.($afdelingSelect[0]),$_.($afdelingSelect[1])) }
        $afdelingen
    }
}

<# Gebruik Join-Object ?? #>
function Hercodeer {
    $Status.AppendText("`r`nStap 5:`tHercodeer datasets met pseudoniemen.")
    "Stap 5:`tHercodeer datasets met pseudoniemen." | Out-File $logfile -Encoding utf8 -Append
    $script:verblijfSelect+="Verblijfsnummer"
    $lengte=$script:patData.length+$script:toedieningData.length
    $ProgressBar.Maximum=$lengte
    $ProgressBar.Step=1
    $ProgressBar.Value=0
    ForEach ($pat in $patData) {
        $pat | Add-Member -Name $verblijfSelect[0] -Value $pathash[(select-Object -ExpandProperty $verblijfSelect[0] -InputObject $pat)] -MemberType NoteProperty -Force
        <# pas geboortedatum aan #>
        $pat.Geboortedatum=$pat.Geboortedatum.substring(3)
        $ProgressBar.PerformStep()
    }
    <# voeg de member Afdeling en toediening toe voor deborah #>
    $script:afdelingslijst = ($script:patData | Get-Afdelingen)
    $script:toedieningSelect+="Verblijfsnummer","Afdeling","Toediening"
    $last_patient=""
    $last_verblijf=""
    $last_afdelingscode=""
    ForEach ($benzopat in $toedieningData) {
        $benzopat | Add-Member -Name $toedieningSelect[0] -Value $pathash[(select-Object -ExpandProperty $verblijfSelect[0] -InputObject $benzopat)] -MemberType NoteProperty -Force
        $dag = (Get-Datum $benzopat.Toedieningsdag ("Ongeldige datum "+$benzopat.Toedieningsdag+" in $bestanden[1] op lijn "+($benzopat.Line+1)+"."))
        if ($benzopat.patient -ne $last_patient -or $benzopat.verblijf -ne $last_verblijf -or $benzopat.afdelingscode -ne $last_afdelingscode) {
            $afdelingsdata = ($patData | `
                Where-Object { $_.Patient -eq $benzopat.Patient -and $_.Verblijf -eq $benzopat.Verblijf -and $_.Afdelingscode -eq $benzopat.Afdelingscode } | `
                ForEach-Object { [pscustomobject]@{ _medontslag = $_._medontslag; _opname = $_._opname ; _ontslag = $_._ontslag; Volgorde = $_.Volgorde ; Verblijfsnummer = $_.Verblijfsnummer } })
            $last_patient=$benzopat.patient
            $last_verblijf=$benzopat.verblijf
            $last_afdelingscode=$benzopat.afdelingscode
        }
        $Verblijfsnummer = ($afdelingsdata | Where-Object {$dag -ge $_._opname -and $dag -le $_._ontslag }).Verblijfsnummer
        $Volgorde = ($afdelingsdata | Where-Object { $dag -ge $_._opname -and $dag -le $_._ontslag }).Volgorde
        Add-Member -Name Verblijfsnummer -Value $Verblijfsnummer -InputObject $benzopat -MemberType NoteProperty -Force
        Add-Member -Name Volgorde -Value $Volgorde -InputObject $benzopat -MemberType NoteProperty -Force
        $afdeling=$afdelingslijst[$benzopat.Afdelingscode]
        Add-Member -Name Afdeling -Value $afdeling -InputObject $benzopat -MemberType NoteProperty -Force
        Add-Member -Name Toediening -Value ($benzopat.aantal + ' ' + $benzopat.eenheid) -InputObject $benzopat -MemberType NoteProperty -Force
        $ProgressBar.PerformStep()
    }
}

<# Stap 6 - Schrijf XML-bestand #>
function Write-ObjectToXML {
    param ([System.Object]$Lijst,[String]$naam)
    $xmlWriter.WriteStartElement($naam+"Table") # <-- Start <$naamTable>
    $Status.AppendText("`r`n`tSchrijf Tabel $naam")
    "`tSchrijf Tabel $naam" | Out-File $logfile -Encoding utf8 -Append
    $ProgressBar.Maximum=$Lijst.length
    $ProgressBar.Value=0
    $Lijst | ForEach-Object  {
        $xmlWriter.WriteStartElement($naam) # <-- Start <$naam>
        $to=$_
        ForEach ($elem in $(Get-Variable -ValueOnly $($naam + "Select"))) {
            $xmlel=select-Object -Property $elem -InputObject $to 
            $xmlWriter.WriteElementString($elem,$xmlel.$elem)
        }
        $ProgressBar.PerformStep() # $syncHash.Window.Dispatcher.invoke([action]{ $syncHash.$ProgressBar.PerformStep() }
        $xmlWriter.WriteEndElement() # <-- End <$naam>
    }
    $xmlWriter.WriteEndElement() # <-- End <$naamTable>
}

Function New-Buffer([string]$Path) {
    $Utf8NoBomEncoding = New-Object System.Text.UTF8Encoding $False
    [System.Int32] $bufferSize = 0x10000 # 64kb buffer

    return (new-object System.IO.StreamWriter -ArgumentList $Path,$false,$Utf8NoBomEncoding,$bufferSize)
}

Function New-XMLWriter ([System.IO.StreamWriter]$Stream) {
    $xmlsettings = New-Object System.Xml.XmlWriterSettings
    $xmlsettings.Indent = $true
    $xmlsettings.IndentChars = "   "
    $XmlWriter = [System.XML.XmlWriter]::Create($Stream, $xmlsettings)

    # Write the XML Declaration and set the XSL
    $xmlWriter.WriteStartDocument()
    return $xmlwriter
}
Function Close-XMLWriter([System.XML.XmlWriter]$xmlwriter) {
    $xmlWriter.WriteEndDocument()
    $xmlWriter.Flush()
    $xmlWriter.Close()
}

function SchrijfXML {
    # Set the File Name Create The Document
    $script:XMLFile="$CDir\"+$userid.Name+"-benzo-"+$regjaar+"-"+$script:numpat+".xml"
    $Status.AppendText("`r`nStap 6:`tXML bestand wordt aangemaakt`r`n`t $XMLFile")
    "Stap 6:`tXML bestand wordt aangemaakt ($XMLFile)" | Out-File $logfile -Encoding utf8 -Append
    $bufferStream = New-Buffer($XMLFile)
    $xmlWriter = New-XMLWriter -Stream $bufferStream

    # Start the Root Element
    $moment=(Get-Date -Format "yyyy-MM-ddTHH:mm")
    $xmlWriter.WriteStartElement("Benzo")
        $xmlwriter.WriteAttributeString("generated", $moment); 
        $xmlwriter.WriteAttributeString("userid", $userid.Name); 
        $xmlwriter.WriteAttributeString("naam", $userid.Value); 
        $xmlwriter.WriteAttributeString("jaar", $regjaar); 
        $xmlwriter.WriteAttributeString("contact", $emailadres); 

    <# schrijf rechtstreeks naar XML #>
    Write-ObjectToXML -Lijst $script:patData -Naam "verblijf"
    Write-ObjectToXML -Lijst $script:toedieningData -Naam "toediening"
    $afdelingen = ($script:afdelingslijst.GetEnumerator() | `
        Select-Object -Property @{N=$afdelingSelect[0];E={$_.Key}},@{N=$afdelingSelect[1];E={$_.Value}} | `
        Sort-Object $afdelingSelect)
    Write-ObjectToXML -Lijst $afdelingen -Naam "afdeling"

    $xmlWriter.WriteEndElement() # <-- End <Benzo> 

    # End, Finalize and close the XML Document
    Close-XMLWriter($xmlWriter)
    $bufferStream.Close()

<# Stap Bereken nu nog de hash en schrijf die ook weg #>
    $Status.AppendText("`r`n`tDe hash wordt berekend en weggeschreven.")
    "`tDe hash wordt berekend en weggeschreven." | Out-File $logfile -Encoding utf8 -Append
    $fhash=(get-FileHash $XMLFile -Algorithm SHA1)
    $hashfile = New-Item $($fhash.path + ".hash") -ItemType "file" -Value $fhash.hash -Force
}

If (LeesBestanden) {
    MaakPseudoniemen
    MaakVolgnummers
    Hercodeer
    SchrijfXML
}
$script:XMLFile
} <# ScriptBlock #>

Function New-Runspace {
    param(
        [System.Management.Automation.ScriptBlock]$Code,
        [Parameter(HelpMessage="Pass arguments to the code")][PSCustomObject[]]$Arguments,
        [Parameter(HelpMessage="Pass variables by name and value")][HashTable]$Params
    )
    $Runspace = [runspacefactory]::CreateRunspace()
    $Runspace.ApartmentState = "STA"
    $Runspace.ThreadOptions = "ReuseThread"
    $Runspace.Open()
    $PSinstance = [powershell]::Create().AddScript($Code)
    ForEach ($arg in $Arguments) { $PSinstance.AddArgument($arg) }
    If ($Params.Count -gt 0) { $Params.GetEnumerator() | ForEach-Object { 
        $Runspace.SessionStateProxy.SetVariable($_.Name,$_.Value) 
    } }
    $PSinstance.Runspace = $Runspace
    $PSinstance
    # $job = $PSinstance.BeginInvoke() om op te starten ...
    # While (-Not $job.IsCompleted) {}  om te wachten tot ze afgelopen is ...
}

Function Close-Runspace([powershell]$PSinstance) {
    <# Catch and display script errors #>
    if ($PSinstance.Streams.Error.Count -gt 0)
    {
        Write-Host "$PSinstance.Streams.Error.Count errors"
        ForEach ($Err in $PSinstance.Streams.Error) {
            $Err.ToString() | Out-File -FilePath $Logfile -Append 
        }
    }
    $PSinstance.Runspace.Close()
    $PSinstance.Runspace.Dispose()
}

#region startApp
Invoke-Expression -Command "$DefMsgBox"
MaakForm
#endregion


